@@ -1,47 +1,101 @@
-/* Copyright (C) 2001 by First Peer, Inc. All rights reserved.
-**
-** Redistribution and use in source and binary forms, with or without
-** modification, are permitted provided that the following conditions
-** are met:
-** 1. Redistributions of source code must retain the above copyright
-**    notice, this list of conditions and the following disclaimer.
-** 2. Redistributions in binary form must reproduce the above copyright
-**    notice, this list of conditions and the following disclaimer in the
-**    documentation and/or other materials provided with the distribution.
-** 3. The name of the author may not be used to endorse or promote products
-**    derived from this software without specific prior written permission. 
-**  
-** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
-** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
-** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
-** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
-** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
-** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
-** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-** SUCH DAMAGE. */
-
 #include "xmlrpc_config.h"
 
+#include <assert.h>
 #include <stddef.h>
 #include <stdlib.h>
 #include <string.h>
 
 #include "xmlrpc-c/base.h"
 #include "xmlrpc-c/base_int.h"
+#include "int.h"
 
 #define KEY_ERROR_BUFFER_SZ (32)
 
 
+
+static uint32_t
+hashStructKey(const char * const key, 
+              size_t       const keyLen) {
+
+    uint32_t hash;
+    size_t i;
+
+    XMLRPC_ASSERT(key != NULL);
+    
+    /* This is the Bernstein hash, optimized for lower case ASCII
+       keys.  Note that the bytes of such a key differ only in their
+       lower 5 bits.
+    */
+    for (hash = 0, i = 0; i < keyLen; ++i)
+        hash = hash + key[i] + (hash << 5);
+
+    return hash;
+}
+
+
+
+static void
+changeMemberValue(xmlrpc_value * const structP,
+                  unsigned int   const mbrIndex,
+                  xmlrpc_value * const newValueP) {
+/*----------------------------------------------------------------------------
+  Change the value of an existing member. (But be careful--the original and
+  new values might be the same object, so watch the order of INCREF and DECREF
+  calls!)
+-----------------------------------------------------------------------------*/
+    _struct_member * const members =
+        XMLRPC_MEMBLOCK_CONTENTS(_struct_member, structP->blockP);
+    _struct_member * const memberP = &members[mbrIndex];
+    xmlrpc_value * const oldValueP = memberP->value;
+
+    /* Juggle our references. */
+    memberP->value = newValueP;
+    xmlrpc_INCREF(memberP->value);
+    xmlrpc_DECREF(oldValueP);
+}
+
+
+
+static void
+addNewMember(xmlrpc_env *   const envP,
+             xmlrpc_value * const structP,
+             xmlrpc_value * const keyvalP,
+             xmlrpc_value * const valueP) {
+/*----------------------------------------------------------------------------
+   Add a new member.  Assume no member already exists with this key.
+-----------------------------------------------------------------------------*/
+    const char * const key =
+        XMLRPC_MEMBLOCK_CONTENTS(char, keyvalP->blockP);
+    size_t const keyLen =
+        XMLRPC_MEMBLOCK_SIZE(char, keyvalP->blockP) - 1;
+
+    _struct_member newMember;
+
+    newMember.keyHash = hashStructKey(key, keyLen);
+    newMember.key     = keyvalP;
+    newMember.value   = valueP;
+
+    XMLRPC_MEMBLOCK_APPEND(_struct_member, envP, structP->blockP,
+                           &newMember, 1);
+
+    if (!envP->fault_occurred) {
+        xmlrpc_INCREF(keyvalP);
+        xmlrpc_INCREF(valueP);
+    }
+}
+
+
+
 void
 xmlrpc_destroyStruct(xmlrpc_value * const structP) {
-
+/*----------------------------------------------------------------------------
+   Dispose of the contents of struct *structP (but not the struct value
+   itself).  The value is not valid after this.
+-----------------------------------------------------------------------------*/
     _struct_member * const members = 
-        XMLRPC_MEMBLOCK_CONTENTS(_struct_member, &structP->_block);
+        XMLRPC_MEMBLOCK_CONTENTS(_struct_member, structP->blockP);
     size_t const size = 
-        XMLRPC_MEMBLOCK_SIZE(_struct_member, &structP->_block);
+        XMLRPC_MEMBLOCK_SIZE(_struct_member, structP->blockP);
 
     unsigned int i;
 
@@ -49,7 +103,7 @@ xmlrpc_destroyStruct(xmlrpc_value * const structP) {
         xmlrpc_DECREF(members[i].key);
         xmlrpc_DECREF(members[i].value);
     }
-    XMLRPC_MEMBLOCK_CLEAN(_struct_member, &structP->_block);
+    XMLRPC_MEMBLOCK_FREE(_struct_member, structP->blockP);
 }
 
 
@@ -76,7 +130,7 @@ xmlrpc_struct_new(xmlrpc_env * const envP) {
     if (!envP->fault_occurred) {
         valP->_type = XMLRPC_TYPE_STRUCT;
 
-        XMLRPC_MEMBLOCK_INIT(_struct_member, envP, &valP->_block, 0);
+        valP->blockP = XMLRPC_MEMBLOCK_NEW(_struct_member, envP, 0);
 
         if (envP->fault_occurred)
             free(valP);
@@ -86,6 +140,66 @@ xmlrpc_struct_new(xmlrpc_env * const envP) {
 
 
 
+xmlrpc_value *
+xmlrpc_struct_new_value(xmlrpc_env *   const envP,
+                        xmlrpc_value * const valueP) {
+
+    xmlrpc_value * structP;
+
+    if (valueP->_type != XMLRPC_TYPE_STRUCT) {
+        xmlrpc_env_set_fault_formatted(envP, XMLRPC_TYPE_ERROR,
+                                       "Value is not a structure.  "
+                                       "It is type #%d", valueP->_type);
+        structP = NULL;
+    } else {
+        size_t const size = 
+            XMLRPC_MEMBLOCK_SIZE(xmlrpc_value *, structP->blockP);
+
+        xmlrpc_createXmlrpcValue(envP, &structP);
+        if (!envP->fault_occurred) {
+            structP->_type = XMLRPC_TYPE_STRUCT;
+
+            structP->blockP = XMLRPC_MEMBLOCK_NEW(xmlrpc_value *, envP, 0);
+
+            if (envP->fault_occurred)
+                free(structP);
+            else {
+                _struct_member * const srcMemberList =
+                    XMLRPC_MEMBLOCK_CONTENTS(_struct_member, valueP->blockP);
+                
+                unsigned int i;
+            
+                for (i = 0; i < size && !envP->fault_occurred; ++i) {
+                    const _struct_member * const thisMemberP =
+                        &srcMemberList[i];
+
+                    xmlrpc_value * keyValP =
+                        xmlrpc_string_new_value(envP, thisMemberP->key);
+                    if (!envP->fault_occurred) {
+                        xmlrpc_value * valueP =
+                            xmlrpc_value_new(envP, thisMemberP->value);
+
+                        if (!envP->fault_occurred) {
+                            addNewMember(envP, structP, keyValP, valueP);
+
+                            xmlrpc_DECREF(valueP);
+                        }
+                        xmlrpc_DECREF(keyValP);
+                    }
+                }
+                if (envP->fault_occurred)
+                    xmlrpc_destroyStruct(structP);
+            }
+
+            if (envP->fault_occurred)
+                free(structP);
+        }
+    }
+    return structP;
+}
+
+
+
 /*=========================================================================
 **  xmlrpc_struct_size
 **=========================================================================
@@ -94,87 +208,75 @@ xmlrpc_struct_new(xmlrpc_env * const envP) {
 */
 
 int 
-xmlrpc_struct_size(xmlrpc_env* env, xmlrpc_value* strct)
-{
+xmlrpc_struct_size(xmlrpc_env *   const envP,
+                   xmlrpc_value * const structP) {
+
     int retval;
 
-    /* Suppress a compiler warning about uninitialized variables. */
-    retval = 0;
+    XMLRPC_ASSERT_ENV_OK(envP);
+    XMLRPC_ASSERT_VALUE_OK(structP);
 
-    XMLRPC_ASSERT_ENV_OK(env);
-    XMLRPC_ASSERT_VALUE_OK(strct);
+    if (structP->_type != XMLRPC_TYPE_STRUCT) {
+        xmlrpc_env_set_fault_formatted(
+            envP, XMLRPC_TYPE_ERROR, "Value is not a struct.  It is type #%d",
+            structP->_type);
+        retval = -1;
+    } else {
+        size_t const size =
+            XMLRPC_MEMBLOCK_SIZE(_struct_member, structP->blockP);
 
-    XMLRPC_TYPE_CHECK(env, strct, XMLRPC_TYPE_STRUCT);
-    retval = XMLRPC_MEMBLOCK_SIZE(_struct_member, &strct->_block);
+        assert((size_t)(int)size == size);
+            /* Because structs are defined to have few enough members */
 
- cleanup:
-    if (env->fault_occurred)
-        return -1;
+        retval = (int)size;
+    }
     return retval;
 }
 
 
 
-/*=========================================================================
-**  get_hash
-**=========================================================================
-**  A mindlessly simple hash function. Please feel free to write something
-**  more clever if this produces bad results.
-*/
-
-static unsigned char 
-get_hash(const char * const key, 
-         size_t       const key_len) {
-
-    unsigned char retval;
-    size_t i;
-
-    XMLRPC_ASSERT(key != NULL);
-    
-    retval = 0;
-    for (i = 0; i < key_len; i++)
-        retval += key[i];
-    return retval;
-}
-
-
-
-/*=========================================================================
-**  find_member
-**=========================================================================
-**  Get the index of the member with the specified key, or -1 if no such
-**  member exists.
-*/
-
-static int 
-find_member(xmlrpc_value * const strctP, 
-            const char *   const key, 
-            size_t         const key_len) {
+static void
+findMember(xmlrpc_value * const structP, 
+           const char *   const key, 
+           size_t         const keyLen,
+           bool *         const foundP,
+           unsigned int * const indexP) {
 
     size_t size, i;
-    unsigned char hash;
-    _struct_member *contents;
-    xmlrpc_value *keyval;
-    char *keystr;
-    size_t keystr_size;
+    uint32_t searchHash;
+    _struct_member * contents;  /* array */
+    bool found;
+    size_t foundIndex;  /* Meaningful only when 'found' is true */
 
-    XMLRPC_ASSERT_VALUE_OK(strctP);
+    XMLRPC_ASSERT_VALUE_OK(structP);
     XMLRPC_ASSERT(key != NULL);
+    foundIndex = 0;  /* defeat used-before-set compiler warning */
 
     /* Look for our key. */
-    hash = get_hash(key, key_len);
-    size = XMLRPC_MEMBLOCK_SIZE(_struct_member, &strctP->_block);
-    contents = XMLRPC_MEMBLOCK_CONTENTS(_struct_member, &strctP->_block);
-    for (i = 0; i < size; i++) {
-        if (contents[i].key_hash == hash) {
-            keyval = contents[i].key;
-            keystr = XMLRPC_MEMBLOCK_CONTENTS(char, &keyval->_block);
-            keystr_size = XMLRPC_MEMBLOCK_SIZE(char, &keyval->_block)-1;
-            if (key_len == keystr_size && memcmp(key, keystr, key_len) == 0)
-                return i;
+    searchHash = hashStructKey(key, keyLen);
+    size = XMLRPC_MEMBLOCK_SIZE(_struct_member, structP->blockP);
+    contents = XMLRPC_MEMBLOCK_CONTENTS(_struct_member, structP->blockP);
+    for (i = 0, found = false; i < size && !found; ++i) {
+        if (contents[i].keyHash == searchHash) {
+            xmlrpc_value * const keyvalP = contents[i].key;
+            const char * const keystr =
+                XMLRPC_MEMBLOCK_CONTENTS(char, keyvalP->blockP);
+            size_t const keystrSize =
+                XMLRPC_MEMBLOCK_SIZE(char, keyvalP->blockP)-1;
+            if (keystrSize == keyLen && memcmp(key, keystr, keyLen) == 0) {
+                found = true;
+                foundIndex = i;
+            }
         }   
     }
-    return -1;
+    if (found) {
+        assert((size_t)(int)foundIndex == foundIndex);
+            /* Definition of structure says it has few enough members */
+
+        if (indexP)
+            *indexP = foundIndex;
+    }
+    *foundP = found;
 }
 
 
@@ -197,25 +299,29 @@ xmlrpc_struct_has_key(xmlrpc_env *   const envP,
 
 int 
 xmlrpc_struct_has_key_n(xmlrpc_env   * const envP,
-                        xmlrpc_value * const strctP,
+                        xmlrpc_value * const structP,
                         const char *   const key, 
-                        size_t         const key_len) {
-    int index;
+                        size_t         const keyLen) {
+    int retval;
 
     /* Suppress a compiler warning about uninitialized variables. */
-    index = 0;
+    retval = 0;
 
     XMLRPC_ASSERT_ENV_OK(envP);
-    XMLRPC_ASSERT_VALUE_OK(strctP);
+    XMLRPC_ASSERT_VALUE_OK(structP);
     XMLRPC_ASSERT(key != NULL);
     
-    XMLRPC_TYPE_CHECK(envP, strctP, XMLRPC_TYPE_STRUCT);
-    index = find_member(strctP, key, key_len);
+    if (structP->_type != XMLRPC_TYPE_STRUCT)
+        xmlrpc_env_set_fault(envP, XMLRPC_TYPE_ERROR,
+                             "Value is not a struct");
+    else {
+        bool found;
 
- cleanup:
-    if (envP->fault_occurred)
-        return 0;
-    return (index >= 0);
+        findMember(structP, key, keyLen, &found, NULL);
+
+        retval = found ? 1 : 0;
+    }
+    return retval;
 }
 
 
@@ -250,15 +356,16 @@ xmlrpc_struct_find_value(xmlrpc_env *    const envP,
             envP, XMLRPC_TYPE_ERROR, "Value is not a struct.  It is type #%d",
             structP->_type);
     else {
-        int index;
+        bool found;
+        unsigned int index;
 
         /* Get our member index. */
-        index = find_member(structP, key, strlen(key));
-        if (index < 0)
+        findMember(structP, key, strlen(key), &found, &index);
+        if (!found)
             *valuePP = NULL;
         else {
             _struct_member * const members =
-                XMLRPC_MEMBLOCK_CONTENTS(_struct_member, &structP->_block);
+                XMLRPC_MEMBLOCK_CONTENTS(_struct_member, structP->blockP);
             *valuePP = members[index].value;
             
             XMLRPC_ASSERT_VALUE_OK(*valuePP);
@@ -294,17 +401,19 @@ xmlrpc_struct_find_value_v(xmlrpc_env *    const envP,
                 "It is type #%d",
                 keyP->_type);
         else {
-            int index;
+            bool found;
+            unsigned int index;
 
             /* Get our member index. */
-            index = find_member(structP, 
-                                XMLRPC_MEMBLOCK_CONTENTS(char, &keyP->_block),
-                                XMLRPC_MEMBLOCK_SIZE(char, &keyP->_block)-1);
-            if (index < 0)
+            findMember(structP, 
+                       XMLRPC_MEMBLOCK_CONTENTS(char, keyP->blockP),
+                       XMLRPC_MEMBLOCK_SIZE(char, keyP->blockP)-1,
+                       &found, &index);
+            if (!found)
                 *valuePP = NULL;
             else {
                 _struct_member * const members =
-                    XMLRPC_MEMBLOCK_CONTENTS(_struct_member, &structP->_block);
+                    XMLRPC_MEMBLOCK_CONTENTS(_struct_member, structP->blockP);
                 *valuePP = members[index].value;
                 
                 XMLRPC_ASSERT_VALUE_OK(*valuePP);
@@ -336,8 +445,8 @@ xmlrpc_struct_read_value_v(xmlrpc_env *    const envP,
         if (*valuePP == NULL) {
             xmlrpc_env_set_fault_formatted(
                 envP, XMLRPC_INDEX_ERROR, "No member of struct has key '%.*s'",
-                (int)XMLRPC_MEMBLOCK_SIZE(char, &keyP->_block),
-                XMLRPC_MEMBLOCK_CONTENTS(char, &keyP->_block));
+                (int)XMLRPC_MEMBLOCK_SIZE(char, keyP->blockP),
+                XMLRPC_MEMBLOCK_CONTENTS(char, keyP->blockP));
         }
     }
 }
@@ -379,9 +488,9 @@ xmlrpc_struct_get_value_n(xmlrpc_env *   const envP,
                           size_t         const keyLen) {
 
     xmlrpc_value * retval;
-    xmlrpc_value * keyP;
+    xmlrpc_value * keyP;  /* 'key' as an XML-RPC string */
     
-    keyP = xmlrpc_build_value(envP, "s#", key, keyLen);
+    keyP = xmlrpc_string_new_lp(envP, keyLen, key);
     if (!envP->fault_occurred) {
         xmlrpc_struct_find_value_v(envP, structP, keyP, &retval);
 
@@ -449,10 +558,9 @@ xmlrpc_struct_set_value_n(xmlrpc_env *    const envP,
             "Type is %d; struct is %d",
             xmlrpc_value_type(strctP), XMLRPC_TYPE_STRUCT);
     else {
-        xmlrpc_value * keyvalP;
+        xmlrpc_value * keyvalP;  /* 'key' as an XML-RPC string */
 
-        /* Get the key as an xmlrpc_value */
-        keyvalP = xmlrpc_build_value(envP, "s#", key, keyLen);
+        keyvalP = xmlrpc_string_new_lp(envP, keyLen, key);
         if (!envP->fault_occurred)
             xmlrpc_struct_set_value_v(envP, strctP, keyvalP, valueP);
 
@@ -464,54 +572,37 @@ xmlrpc_struct_set_value_n(xmlrpc_env *    const envP,
 
 void 
 xmlrpc_struct_set_value_v(xmlrpc_env *   const envP,
-                          xmlrpc_value * const strctP,
+                          xmlrpc_value * const structP,
                           xmlrpc_value * const keyvalP,
                           xmlrpc_value * const valueP) {
 
-    char *key;
-    size_t key_len;
-    int index;
-    _struct_member *members, *member, new_member;
-    xmlrpc_value *old_value;
-
     XMLRPC_ASSERT_ENV_OK(envP);
-    XMLRPC_ASSERT_VALUE_OK(strctP);
+    XMLRPC_ASSERT_VALUE_OK(structP);
     XMLRPC_ASSERT_VALUE_OK(keyvalP);
     XMLRPC_ASSERT_VALUE_OK(valueP);
 
-    XMLRPC_TYPE_CHECK(envP, strctP, XMLRPC_TYPE_STRUCT);
-    XMLRPC_TYPE_CHECK(envP, keyvalP, XMLRPC_TYPE_STRING);
+    if (structP->_type != XMLRPC_TYPE_STRUCT)
+        xmlrpc_env_set_fault(envP, XMLRPC_TYPE_ERROR,
+                             "Value is not a struct");
+    else if (keyvalP->_type != XMLRPC_TYPE_STRING)
+        xmlrpc_env_set_fault(envP, XMLRPC_TYPE_ERROR,
+                             "Key value is not a string");
+    else {
+        const char * const key =
+            XMLRPC_MEMBLOCK_CONTENTS(char, keyvalP->blockP);
+        size_t const keyLen =
+            XMLRPC_MEMBLOCK_SIZE(char, keyvalP->blockP) - 1;
 
-    key = XMLRPC_MEMBLOCK_CONTENTS(char, &keyvalP->_block);
-    key_len = XMLRPC_MEMBLOCK_SIZE(char, &keyvalP->_block) - 1;
-    index = find_member(strctP, key, key_len);
+        bool found;
+        unsigned int index;
 
-    if (index >= 0) {
-        /* Change the value of an existing member. (But be careful--the
-        ** original and new values might be the same object, so watch
-        ** the order of INCREF and DECREF calls!) */
-        members = XMLRPC_MEMBLOCK_CONTENTS(_struct_member, &strctP->_block);
-        member = &members[index];
+        findMember(structP, key, keyLen, &found, &index);
 
-        /* Juggle our references. */
-        old_value = member->value;
-        member->value = valueP;
-        xmlrpc_INCREF(member->value);
-        xmlrpc_DECREF(old_value);
-    } else {
-        /* Add a new member. */
-        new_member.key_hash = get_hash(key, key_len);
-        new_member.key      = keyvalP;
-        new_member.value    = valueP;
-        XMLRPC_MEMBLOCK_APPEND(_struct_member, envP, &strctP->_block,
-                               &new_member, 1);
-        XMLRPC_FAIL_IF_FAULT(envP);
-        xmlrpc_INCREF(keyvalP);
-        xmlrpc_INCREF(valueP);
+        if (found)
+            changeMemberValue(structP, index, valueP);
+        else
+            addNewMember(envP, structP, keyvalP, valueP);
     }
-
-cleanup:
-    return;
 }
 
 
@@ -538,9 +629,9 @@ xmlrpc_struct_read_member(xmlrpc_env *    const envP,
             "of something that is not a struct");
     else {
         _struct_member * const members =
-            XMLRPC_MEMBLOCK_CONTENTS(_struct_member, &structP->_block);
+            XMLRPC_MEMBLOCK_CONTENTS(_struct_member, structP->blockP);
         size_t const size = 
-            XMLRPC_MEMBLOCK_SIZE(_struct_member, &structP->_block);
+            XMLRPC_MEMBLOCK_SIZE(_struct_member, structP->blockP);
 
         if (index >= size)
             xmlrpc_env_set_fault_formatted(
@@ -590,3 +681,31 @@ xmlrpc_struct_get_key_and_value(xmlrpc_env *    const envP,
         *valueP = NULL;
     }
 }
+
+
+
+/* Copyright (C) 2001 by First Peer, Inc. All rights reserved.
+**
+** Redistribution and use in source and binary forms, with or without
+** modification, are permitted provided that the following conditions
+** are met:
+** 1. Redistributions of source code must retain the above copyright
+**    notice, this list of conditions and the following disclaimer.
+** 2. Redistributions in binary form must reproduce the above copyright
+**    notice, this list of conditions and the following disclaimer in the
+**    documentation and/or other materials provided with the distribution.
+** 3. The name of the author may not be used to endorse or promote products
+**    derived from this software without specific prior written permission. 
+**  
+** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
+** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
+** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+** SUCH DAMAGE. */
+
